"""This file has functions that handle the user cli commands"""

import subprocess
import shutil
from src.pkg import utils
from src.pkg.constants import COMPOSE_USERS_YML

def _build_config_mapping(user_config, resources):
    """Build the mapping for config substitution.

    Args:
        user_config: Dict with keys 'username', 'path', optionally 'server'
        resources: Dict with resource limits

    Returns:
        Mapping dict for config substitution
    """
    mapping = {
        "${DTAAS_DIR}": user_config["path"],
        "${username}": user_config["username"],
        "${shm_size}": str(resources["shm_size"]),
        "${cpus}": str(resources["cpus"]),
        "${mem_limit}": str(resources["mem_limit"]),
        "${pids_limit}": str(resources["pids_limit"])
    }
    if user_config.get("server") is not None:
        mapping["${SERVER_DNS}"] = user_config["server"]
    return mapping


def _load_template(server):
    """Load the appropriate template based on server type."""
    if server == utils.LOCALHOST_SERVER:
        return utils.import_yaml("users.local.yml")
    return utils.import_yaml("users.server.yml")


def get_compose_config(username, config):
    """Makes and returns the config for the user

    Args:
        username: Username for the config
        config: Dict with 'server', 'path', 'resources' keys

    Returns:
        Tuple of (user config dict, error if any)
    """
    try:
        template, err = _load_template(config["server"])
        utils.check_error(err)
        user_config = {
            "username": username,
            "path": config["path"],
            "server": config["server"] if config["server"] != utils.LOCALHOST_SERVER else None
        }
        mapping = _build_config_mapping(user_config, config["resources"])
        result, err = utils.replace_all(template, mapping)
        utils.check_error(err)
    except Exception as e:
        return None, e
    return result, None


def create_user_files(users, file_path):
    """Creates all the users' workspace directories"""
    for username in users:
        shutil.copytree(
            file_path + "/template", file_path + "/" + username, dirs_exist_ok=True
        )


def add_users_to_compose(users, compose, config):
    """Adds all the users config to the compose dictionary
    Args:
        users: List of usernames
        compose: Compose dict to update
        config: Dict with 'server', 'path', 'resources' keys
    """
    for username in users:
        user_conf, err = get_compose_config(username, config)
        if err is not None:
            return err
        compose["services"][username] = user_conf
    return None


def start_user_containers(users):
    """Starts all the user containers in the 'users' list"""
    cmd = "docker compose -f compose.users.yml up -d"
    err = run_command_for_containers(cmd, users)
    return err


def stop_user_containers(users):
    """Stops all the user containers in the 'users' list"""
    cmd = "docker compose -f compose.users.yml down"
    err = run_command_for_containers(cmd, users)
    return err


def run_command_for_containers(command, containers):
    """Runs the given docker command for the given containers"""
    cmd = [command]
    for name in containers:
        cmd.append(name)
    cmd_str = " ".join(cmd)
    result = subprocess.run(cmd_str, shell=True, check=False)
    if result.returncode != 0:
        return Exception(f"failed to run '{cmd_str}' command")
    return None


def _setup_compose_structure(compose):
    """Ensure compose has required structure for services."""
    if "version" not in compose:
        compose["version"] = "3"
    if "services" not in compose:
        compose["services"] = {}
    if "networks" not in compose:
        compose["networks"] = {"users": {"name": "dtaas-users", "external": True}}


def _get_add_users_config(config_obj):
    """Retrieve configuration needed for adding users."""
    user_list, err = config_obj.get_add_users_list()
    utils.check_error(err)
    server, err = config_obj.get_server_dns()
    utils.check_error(err)
    path, err = config_obj.get_path()
    utils.check_error(err)
    resources, err = config_obj.get_resource_limits()
    utils.check_error(err)
    return user_list, server, path, resources


def _finalize_compose(compose):
    """Export and start user containers."""
    err = utils.export_yaml(compose, COMPOSE_USERS_YML)
    utils.check_error(err)
    users_list = list(compose["services"].keys())
    err = start_user_containers(users_list)
    utils.check_error(err)


def add_users(config_obj):
    """add cli command handler"""
    try:
        compose, err = utils.import_yaml(COMPOSE_USERS_YML)
        utils.check_error(err)
        user_list, server, path, resources = _get_add_users_config(config_obj)
    except Exception as e:
        return e

    _setup_compose_structure(compose)

    try:
        create_user_files(user_list, path + "/files")
        config = {"server": server, "path": path, "resources": resources}
        err = add_users_to_compose(user_list, compose, config)
        utils.check_error(err)
        _finalize_compose(compose)
    except Exception as e:
        return e

    return None


def _remove_users_from_compose(compose, user_list):
    """Remove users from compose configuration."""
    for username in user_list:
        if "services" in compose and username in compose["services"]:
            del compose["services"][username]


def delete_user(config_obj):
    """delete cli command handler"""
    try:
        compose, err = utils.import_yaml(COMPOSE_USERS_YML)
        utils.check_error(err)
        user_list, err = config_obj.get_delete_users_list()
        utils.check_error(err)
        err = stop_user_containers(user_list)
        utils.check_error(err)
        _remove_users_from_compose(compose, user_list)
        err = utils.export_yaml(compose, COMPOSE_USERS_YML)
        utils.check_error(err)
    except Exception as e:
        return e

    return None



"This file has generic helper functions and variables for dtaas cli"

import yaml
import tomlkit

LOCALHOST_SERVER = "localhost"


def import_yaml(filename):
    """This function is used to import a yaml file safely"""
    config = {}
    try:
        with open(filename, "r") as file:
            config = yaml.safe_load(file)
    except FileNotFoundError:
        return {}, None
    except Exception as err:
        return None, Exception(
            f"Error while getting yaml file: {filename}, " + str(err)
        )
    return config if config is not None else {}, None


def export_yaml(data, filename):
    """This function is used to export to a yaml file safely"""
    try:
        with open(filename, "w") as file:
            yaml.safe_dump(
                data,
                file,
                sort_keys=False,
                default_flow_style=False,
                allow_unicode=True,
                indent=2,
            )
    except Exception as err:
        return Exception(f"Error while writing yaml to file: {filename}, " + str(err))
    return None


def import_toml(filename):
    """This function is used to import a toml file safely"""
    try:
        with open(filename, "r") as file:
            config = tomlkit.load(file)
    except Exception as err:
        return None, Exception(
            f"Error while getting toml file: {filename}, " + str(err)
        )
    return config, None


def replace_all(obj, mapping):
    """This function is used to replace all placeholders with values in a nested object"""
    if isinstance(obj, str):
        obj, err = replace_string(obj, mapping)
        return obj, err

    if isinstance(obj, list):
        obj, err = replace_list(obj, mapping)
        return obj, err

    if isinstance(obj, dict):
        obj, err = replace_dict(obj, mapping)
        return obj, err

    return None, Exception("Config substition failed: Object format not valid")


def replace_string(s, mapping):
    """Replaces all placeholders in the string with values from the mapping"""
    for key in mapping:
        s = s.replace(key, mapping[key])
    return s, None


def replace_list(arr, mapping):
    """Replaces all placeholders in the list with values from the mapping"""
    for ind, val in enumerate(arr):
        arr[ind], err = replace_all(val, mapping)
        if err is not None:
            return None, err
    return arr, None


def replace_dict(dictionary, mapping):
    """Replaces all placeholders in the dictionary with values from the mapping"""
    for key in dictionary:
        if not isinstance(key, str):
            return None, Exception("Config substitution failed: Key is not a string")
        dictionary[key], err = replace_all(dictionary[key], mapping)
        if err is not None:
            return None, err
    return dictionary, None


def check_error(err):
    """Checks if error is not None and raises it"""
    if err is not None:
        raise err



"""This file supports the DTaaS config class"""

import click
from src.pkg import utils


class Config:
    """The Config class for DTaaS"""

    def __init__(self):
        config, err = utils.import_toml("dtaas.toml")
        if err is not None:
            raise click.ClickException("config initialisation failed: " + str(err))
        self.data = config

    def get_config(self):
        """Gets the config dictionary"""
        if self.data is not None:
            return self.data, None
        return None, Exception("Config not initialised")

    def get_from_config(self, key):
        """Gets the specific key from config"""
        conf, err = self.get_config()
        if err is not None:
            return None, err

        if key not in conf:
            return None, Exception(f"Config file error: Missing {key} tag")
        return conf[key], None

    def get_common(self):
        """Gets the 'common' section of config"""
        common, err = self.get_from_config("common")
        return common, err

    def get_string_from_common(self, key):
        """Gets the specific string key from config.common"""
        conf_common, err = self.get_common()
        if err is not None:
            return None, err

        if key not in conf_common or conf_common[key] == "":
            return None, Exception(
                f"Config file error: config.common.{key} not set in TOML"
            )
        return str(conf_common[key]), None

    def get_users(self):
        """Gets the 'users' section of config"""
        users, err = self.get_from_config("users")
        return users, err


    def get_string_list_from_users(self, key):
        """Gets the specific key as a list of strings from config.users"""
        err = None
        strings_list = None

        conf_users, err = self.get_users()
        if err is not None:
            return None, err

        try:
            strings_list = [ str(x) for x in conf_users[key]]
            if len(strings_list) == 0:
                strings_list = None
                raise ValueError
        except KeyError:
            err = Exception(f"Config file error: No {key} list in 'users' tag")
        except ValueError:
            err = Exception(f'Config file error: users.{key} list is empty')

        return strings_list, err

    def get_path(self):
        """Gets the 'path' from config.common"""
        path, err = self.get_string_from_common("path")
        return path, err

    def get_server_dns(self):
        """Gets the 'server-dns' from config.common"""
        server, err = self.get_string_from_common("server-dns")
        return server, err

    def get_add_users_list(self):
        """Gets the 'add' list from config.users"""
        add_users_list, err = self.get_string_list_from_users("add")
        return add_users_list, err

    def get_delete_users_list(self):
        """Gets the 'delete' list from config.users"""
        delete_users_list, err = self.get_string_list_from_users("delete")
        return delete_users_list, err

    def get_resource_limits(self):
        """Gets the default resourse limits"""
        conf_common, err = self.get_common()
        if err is not None:
            return None, err
        # It's assumed that resources is given in dtaas.toml
        resources = conf_common.get("resources", None)
        if resources is None:
            err = Exception("Config file error: Missing default resources limits")
            return None, err
        return resources, None



"""This file defines all cli entrypoints for DTaaS"""

import click
from .pkg import config as configPkg
from .pkg import users as userPkg


### Groups
@click.group()
def dtaas():
    """all commands to help with Digital Twins as a Service"""
    return


@dtaas.group()
def admin():
    "administrative commands for DTaaS"
    return


@admin.group()
def user():
    """user management commands"""
    return


#### user group commands
@user.command()
def add():
    """
    add a list of users to DTaaS at once\n
    Specify the list in dtaas.toml [users].add\n
    """

    config_obj = configPkg.Config()

    err = userPkg.add_users(config_obj)
    if err is not None:
        raise click.ClickException("Error while adding users: " + str(err))
    click.echo("Users added successfully")


@user.command()
def delete():
    """
    removes the USERNAME user from DTaaS\n
    Specify the users in dtaas.toml [users].delete\n
    """

    config_obj = configPkg.Config()

    err = userPkg.delete_user(config_obj)
    if err is not None:
        raise click.ClickException("Error while deleting users: " + str(err))
    click.echo("User deleted successfully")
